﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Common;
using HeuristicLab.PDPSimulation.Operators;
using HeuristicLab.Problems.VehicleRouting;
using HeuristicLab.Problems.VehicleRouting.Interfaces;
using HeuristicLab.Parameters;
using HeuristicLab.Data;
using HeuristicLab.Problems.VehicleRouting.ProblemInstances;

namespace HeuristicLab.PDPSimulation {
  [Item("DynamicWaitingStrategy", @"A pickup and delivery waiting strategy. It has been implemented as described in:
Mitrović-Minić, Snežana, and Gilbert Laporte. Waiting strategies for the dynamic pickup and delivery problem with time windows. Transportation Research Part B: Methodological 38.7 (2004): 635-655.")]
  [StorableClass]
  public class DynamicWaitingStrategy : WaitingStrategy {
    public ValueParameter<DoubleValue> MaxZoneSizeParameter {
      get { return (ValueParameter<DoubleValue>)Parameters["MaxZoneSize"]; }
    }

    public ValueParameter<BoolValue> DistributeTimeParameter {
      get { return (ValueParameter<BoolValue>)Parameters["DistributeTime"]; }
    }

    public DynamicWaitingStrategy()
      : base() {
        Parameters.Add(new ValueParameter<DoubleValue>("MaxZoneSize", "The maximum zone size.", new DoubleValue(150)));
        Parameters.Add(new ValueParameter<BoolValue>("DistributeTime", "Indicates if the waiting time should be distributed proportionally.", new BoolValue(true)));
    }
    [StorableConstructor]
    protected DynamicWaitingStrategy(bool deserializing) : base(deserializing) { }
    protected DynamicWaitingStrategy(DynamicWaitingStrategy original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new DynamicWaitingStrategy(this, cloner);
    }

    protected struct Stop {
      public int Customer { get; set; }
      public double CoordX { get; set; }
      public double CoordY { get; set; }
      public double ArrivalTime { get; set; }
      public double LeaveTime { get; set; }
    }

    private double CalculateBase(List<Stop> stops) {
      double baseSize = 0;

      double minX = stops.Min(s => s.CoordX);
      double minY = stops.Min(s => s.CoordY);
      double maxX = stops.Max(s => s.CoordX);
      double maxY = stops.Max(s => s.CoordY);

      baseSize = (maxX - minX) * (maxY - minY);

      return baseSize;
    }

    protected Dictionary<int, List<Stop>> GetZones(List<Stop> stops) {
      double maxZoneSize = MaxZoneSizeParameter.Value.Value;
      
      Dictionary<int, List<Stop>> zones = new Dictionary<int, List<Stop>>();
      int current = 0;
      List<Stop> currentStops = new List<Stop>();
      zones[current] = currentStops;

      for (int i = 0; i < stops.Count; i++) {
        currentStops.Add(stops[i]);
        double baseSize = CalculateBase(currentStops);

        if (baseSize >= maxZoneSize) {
          currentStops.Remove(stops[i]);

          current = current + 1;
          currentStops = new List<Stop>();
          zones[current] = currentStops;

          currentStops.Add(stops[i]);
        }
      }

      return zones;
    }

    private double GetZoneDuration(List<Stop> stops) {
      double duration = 0;

      if (stops.Count > 0) {
        duration = stops.Last().LeaveTime - stops.First().ArrivalTime;
      }

      return duration;      
    }

    private List<Stop> GetStops(DynPDPProblemInstance instance, IVRPEncoding solution, Tour tour) {
      List<Stop> stops = new List<Stop>();
      Stop stop;

      int vehicleIdx = solution.GetVehicleAssignment(solution.GetTourIndex(tour));
      var vehicle = instance.Simulation.GetVehicle(instance.Wrapper.GetVehicle(vehicleIdx));

      HashSet<Guid> pickedUp = new HashSet<Guid>();
      foreach (var orderId in vehicle.Orders) {
        var order = instance.Simulation.GetOrder(orderId);
        double coordX, coordY, arrivalTime, leaveTime;

        if (pickedUp.Contains(orderId)) {
          coordX = order.DeliveryXCoord;
          coordY = order.DeliveryYCoord;

          if (!vehicle.DeliveredOrders.Contains(orderId))
            arrivalTime = vehicle.ReadyTime - order.DeliveringServiceTime;
          else
            arrivalTime = order.DeliveryTime - order.DeliveryServiceTime;
          leaveTime = arrivalTime + order.DeliveryServiceTime;
        } else {
          pickedUp.Add(orderId);

          coordX = order.PickupXCoord;
          coordY = order.PickupYCoord;
          if (!vehicle.PickedUpOrders.Contains(orderId) && !vehicle.DeliveredOrders.Contains(orderId))
            arrivalTime = vehicle.ReadyTime - order.PickingUpServiceTime;
          else
            arrivalTime = order.PickupTime - order.PickupServiceTime;
          leaveTime = arrivalTime + order.PickupServiceTime;
        }

        stop = new Stop();
        stop.Customer = -1;
        stop.CoordX = coordX;
        stop.CoordY = coordY;
        stop.ArrivalTime = arrivalTime;
        stop.LeaveTime = leaveTime;

        stops.Add(stop);
      }

      VRPEvaluation eval = instance.EvaluateTour(tour, solution);
      TourInsertionInfo info = eval.InsertionInfo.GetTourInsertionInfo(0);
      for (int i = 0; i < tour.Stops.Count; i++) {
        CVRPTWInsertionInfo stopInfo = (CVRPTWInsertionInfo)info.GetStopInsertionInfo(i + 1);

        stop = new Stop();
        stop.Customer = tour.Stops[i];
        stop.CoordX = instance.GetCoordinates(tour.Stops[i])[0];
        stop.CoordY = instance.GetCoordinates(tour.Stops[i])[0];
        stop.ArrivalTime = stopInfo.ArrivalTime + instance.Simulation.SimulationTime;
        stop.LeaveTime = stopInfo.LeaveTime + instance.Simulation.SimulationTime;

        stops.Add(stop);
      }

      return stops;
    }

    public override double GetWaitingTime(double time, DynPDPProblemInstance instance, IVRPEncoding solution, Tour tour, int stop) {
      var stops = GetStops(instance, solution, tour);
      var zones = GetZones(stops);

      Dictionary<int, int> stopZones = new Dictionary<int, int>();
      foreach (var kvp in zones) {
        foreach (var s in kvp.Value) {
          if (s.Customer >= 0)
            stopZones[s.Customer] = kvp.Key;
        }
      }

      var currentZone = stopZones[tour.Stops[stop]];
      int prevZone = -1;
      if (stop > 0)
        prevZone = stopZones[tour.Stops[stop - 1]];
      else {
        int curIdx = stops.IndexOf(stops.First(s => s.Customer == tour.Stops[stop]));
        if (curIdx == 0)
          prevZone = currentZone;
        else {
          Stop prevStop = stops[curIdx - 1];
          prevZone = zones.First(kvp => kvp.Value.Contains(prevStop)).Key;
        }
      }

      if (currentZone != prevZone) {
        double slack = WaitFirstWaitingStrategy.GetSlack(time, instance, solution, tour, stop);

        if (DistributeTimeParameter.Value.Value) {
          double zoneTime = GetZoneDuration(zones[prevZone]);
          
          double tourTime = 0;
          foreach (var kvp in zones.Where(z => z.Key >= prevZone)) {
            tourTime += GetZoneDuration(zones[kvp.Key]);
          }

          if (tourTime > 0)
            slack *= zoneTime / tourTime;
        }
      
        return slack;
      } else {
        return 0;      
      }
    }
  }
}
