﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Common;
using HeuristicLab.PDPSimulation.Operators;
using HeuristicLab.Problems.VehicleRouting;
using HeuristicLab.Problems.VehicleRouting.Interfaces;
using HeuristicLab.PDPSimulation.DomainModel;
using HeuristicLab.Parameters;
using HeuristicLab.Data;
using HeuristicLab.PDPSimulation.DistanceMeasures;

namespace HeuristicLab.PDPSimulation {
  [Item("IntensityWaitingStrategy", "A novel pickup and delivery waiting strategy based on an intensity measure.")]
  [StorableClass]
  public class AdvancedWaitingStrategy : WaitingStrategy {
    public ValueParameter<DoubleMatrix> StrategyVectorParameter {
      get { return (ValueParameter<DoubleMatrix>)Parameters["StrategyVector"]; }
    }

    public OptionalValueParameter<HistoricalData> HistoricalDataParameter {
      get { return (OptionalValueParameter<HistoricalData>)Parameters["HistoricalData"]; }
    }
    
    public AdvancedWaitingStrategy()
      : base() {
        Parameters.Add(new ValueParameter<DoubleMatrix>("StrategyVector", "Waiting strategy vector.", new DoubleMatrix(4, 1)));
        Parameters.Add(new OptionalValueParameter<HistoricalData>("HistoricalData", "The historical data.", null));
    }
    [StorableConstructor]
    protected AdvancedWaitingStrategy(bool deserializing) : base(deserializing) { }
    protected AdvancedWaitingStrategy(AdvancedWaitingStrategy original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new AdvancedWaitingStrategy(this, cloner);
    }

    public override void SetScenario(PickupDeliveryScenario scenario) {
      base.SetScenario(scenario);
    }

    public override double GetWaitingTime(double time, DynPDPProblemInstance instance, IVRPEncoding solution, Tour tour, int stop) {
      int vehicle = solution.GetVehicleAssignment(solution.GetTourIndex(tour));
      bool moving = instance.VehicleStates[vehicle] == DomainModel.VehicleState.Moving;
      if ((!moving && stop != 0) || (moving && stop != 1))
        return 0;

      stop -= 1;

      int depot = -1;
      int customer;
      double[] customerCoords;
      int next;
      double[] nextCoords;
      if (stop >= 0) {
        customer = tour.Stops[stop];

        customerCoords = instance.GetCoordinates(customer);
        if (stop < tour.Stops.Count - 1) {
          next = tour.Stops[stop + 1];

          nextCoords = instance.GetCoordinates(next);
        } else {
          next = 0;
          depot = instance.GetDepot(customer, solution);

          nextCoords = new double[2];
          nextCoords[0] = instance.Coordinates[depot, 0];
          nextCoords[1] = instance.Coordinates[depot, 1];
        }
      } else {
        customer = 0;
        next = tour.Stops[stop + 1];
        depot = instance.GetDepot(next, solution);

        customerCoords = new double[2];
        customerCoords[0] = instance.Coordinates[depot, 0];
        customerCoords[1] = instance.Coordinates[depot, 1];
        nextCoords = instance.GetCoordinates(next);
      }

      DoubleMatrix strategyVector = StrategyVectorParameter.Value;
      double a = strategyVector[0, 0];
      double b = strategyVector[1, 0];
      double c = strategyVector[2, 0];

      double maxV = (a > 0 ? a : 0) + (b > 0 ? b : 0) + (c > 0 ? c : 0);
      double minV = (a < 0 ? a : 0) + (b < 0 ? b : 0) + (c < 0 ? c : 0);    
      
      double maxTime = scenario.VehicleDueTimesParameter.Value.Max();
      double maxDistance = Math.Sqrt(Math.Pow(scenario.OrderRangeX(), 2) + Math.Pow(scenario.OrderRangeY(), 2));

      double threshold = minV + strategyVector[3, 0] * (maxV - minV);

      double dist = instance.GetDistance(customer, next, solution);
      double distance = dist / maxDistance;
      double buffer = WaitFirstWaitingStrategy.GetSlack(time, instance, solution, tour, stop + 1);
      double slack = Math.Min(1.0,  buffer / maxTime);

      double regionIntensity = GetRegionIntensity(time, customerCoords[0], customerCoords[1], time, maxTime);
      
      double arrivalTime = time;
      arrivalTime += dist;
      int idx = 0;
      if (next == 0)
        idx = depot;
      else
        idx = next + instance.Depots.Value - 1;
      if (arrivalTime < instance.ReadyTime[idx])
        arrivalTime = instance.ReadyTime[idx];
      if (next != 0)
        arrivalTime += instance.ServiceTime[next - 1];
      double nextRegionIntensity = GetRegionIntensity(time, nextCoords[0], nextCoords[1], arrivalTime, maxTime);

      double intensity = 1;
      if (regionIntensity + nextRegionIntensity > 0)
        intensity = regionIntensity / (regionIntensity + nextRegionIntensity);

      if(next != 0)
        arrivalTime -= instance.ServiceTime[next - 1];

      double v = a * (arrivalTime - time) / maxTime + b * intensity + c * slack;

      if (v >= threshold) {
        if (maxV - threshold == 0)
          return buffer;
        else
          return Math.Min(buffer, buffer * ((v - threshold) / (maxV - threshold)));
      } else
        return 0;
    }

    public static double GetIntensity(HistoricalData histData, DistanceMeasure distanceMeasure, double x, double y, double time, double arrivalTime, double maxTime) {
      IEnumerable<Order> allOrders = histData.GetOrders(time, double.MaxValue, false).SelectMany(o => o);

      double intensity = 0;
      foreach (Order order in allOrders) {
        double distance = distanceMeasure.GetDistance(x, y, order.PickupXCoord, order.PickupYCoord);

        double waitingTime = Math.Max(0, order.PickupReadyTime - (arrivalTime + distance));
        double orderIntensity = 0;

        if ((arrivalTime + distance) < order.PickupDueTime && maxTime - time > 0)
          orderIntensity = Math.Max(0, 1.0 - (distance + waitingTime) / (maxTime - time));

        intensity += orderIntensity;
      }

      return intensity;
    }

    private double GetRegionIntensity(double time, double x, double y, double arrivalTime, double maxTime) {
      return GetIntensity(HistoricalDataParameter.Value, scenario.DistanceMeasure, x, y, time, arrivalTime, maxTime);
    }
  }
}
