﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using HeuristicLab.Core;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;
using HeuristicLab.Common;
using HeuristicLab.PDPSimulation.Operators;
using HeuristicLab.Problems.VehicleRouting;
using HeuristicLab.Problems.VehicleRouting.Interfaces;
using HeuristicLab.Parameters;
using HeuristicLab.Data;
using HeuristicLab.Problems.VehicleRouting.ProblemInstances;

namespace HeuristicLab.PDPSimulation {
  [Item("StochasticWaitingStrategy", @"A pickup and delivery waiting strategy. It has been implemented as described in: 
Ichoua, Soumia, Michel Gendreau, and Jean-Yves Potvin. Exploiting knowledge about future demands for real-time vehicle dispatching. Transportation Science 40.2 (2006): 211-225.")]
  [StorableClass]
  public class StochasticWaitingStrategy : WaitingStrategy {
    public ValueParameter<DoubleValue> MinDistanceParameter {
      get { return (ValueParameter<DoubleValue>)Parameters["MinDistance"]; }
    }

    public ValueParameter<DoubleValue> MaxWaitTimeParameter {
      get { return (ValueParameter<DoubleValue>)Parameters["MaxWaitTime"]; }
    }

    public ValueParameter<DoubleValue> ThresholdParameter {
      get { return (ValueParameter<DoubleValue>)Parameters["Threshold"]; }
    }

    public ValueParameter<DoubleValue> NeighborhoodSizeParameter {
      get { return (ValueParameter<DoubleValue>)Parameters["NeighborhoodSize"]; }
    }

    public ValueParameter<DoubleValue> MaxVehiclesParameter {
      get { return (ValueParameter<DoubleValue>)Parameters["MaxVehicles"]; }
    }

    public StochasticWaitingStrategy()
      : base() {
        Parameters.Add(new ValueParameter<DoubleValue>("MinDistance", "The minimum distance.", new DoubleValue(0.2)));
        Parameters.Add(new ValueParameter<DoubleValue>("MaxWaitTime", "The maximum wait time.", new DoubleValue(0.1)));
        Parameters.Add(new ValueParameter<DoubleValue>("Threshold", "The threshold.", new DoubleValue(0.8)));
        Parameters.Add(new ValueParameter<DoubleValue>("NeighborhoodSize", "The neighborhood size.", new DoubleValue(0.2)));
        Parameters.Add(new ValueParameter<DoubleValue>("MaxVehicles", "The max number of vehicles.", new DoubleValue(0.5)));
    }
    [StorableConstructor]
    protected StochasticWaitingStrategy(bool deserializing) : base(deserializing) { }
    protected StochasticWaitingStrategy(StochasticWaitingStrategy original, Cloner cloner)
      : base(original, cloner) {
    }
    public override IDeepCloneable Clone(Cloner cloner) {
      return new StochasticWaitingStrategy(this, cloner);
    }

    private double GetAverageDistance(HistoricalData data, DistanceMeasures.DistanceMeasure dist) {
      double avgDistance = 0;

      var orders = data.GetOrders(0, double.MaxValue).SelectMany(l => l).ToList();

      int count = 0;
      double sum = 0;
      for (int i = 0; i < orders.Count; i++) {
        for (int j = i + 1; j < orders.Count; j++) {
          sum += dist.GetDistance(orders[i].PickupXCoord, orders[i].PickupYCoord, orders[j].PickupXCoord, orders[j].PickupYCoord);
          count++;

          sum += dist.GetDistance(orders[i].DeliveryXCoord, orders[i].DeliveryYCoord, orders[j].DeliveryXCoord, orders[j].DeliveryYCoord);
          count++;
        }
      }

      if (count > 0)
        avgDistance = sum / (double)count;

      return avgDistance;
    }

    private double GetProbability(double time, double waitingTime, double posX, double posY, HistoricalDataForecastModel forecastModel) {
      var intensityInfo = forecastModel.GetIntensityInformation(Resource.Requests) as DiscreteIntensityInformation;
      double neighborhoodSize = NeighborhoodSizeParameter.Value.Value * forecastModel.GetWidth();

      int m1 = RegionHelper.GetTimeSlice(time, forecastModel.TimeSlicesParameter.Value.Value, forecastModel.GetHorizon(), false);
      int m2 = RegionHelper.GetTimeSlice(time + waitingTime, forecastModel.TimeSlicesParameter.Value.Value, forecastModel.GetHorizon(), false);

      double probability = 1;
      for (int t = m1; t <= m2; t++) {
        double totalIntensity = 0;

        var upperLeft = RegionHelper.GetGeographicSlice(posX - neighborhoodSize / 2,
           posY - neighborhoodSize / 2, forecastModel.GeographicSlicesParameter.Value.Value,
           forecastModel.GetWidth(), forecastModel.GetHeight(), false);

        var lowerRight = RegionHelper.GetGeographicSlice(posX + neighborhoodSize / 2,
           posY + neighborhoodSize / 2, forecastModel.GeographicSlicesParameter.Value.Value,
           forecastModel.GetWidth(), forecastModel.GetHeight(), false);

        for (int x = upperLeft.Item1; x <= lowerRight.Item1; x++) {
          for (int y = upperLeft.Item2; y <= lowerRight.Item2; y++) {
            var regionIntensityInfo = intensityInfo.GetIntensity(t, new Tuple<int, int>(x, y));
            if (regionIntensityInfo != null) {
              double intensity = 0;
              double confidence = 0;
              regionIntensityInfo.GetIntensity(time, out intensity, out confidence);

              if (intensity > 0) {
                //normalize acc. to region overlap
                double overlap = 1.0;

                if ((x == upperLeft.Item1) || (x == lowerRight.Item1) || (y == upperLeft.Item2) || (y == lowerRight.Item2)) {
                  double width = forecastModel.GetWidth() / forecastModel.GeographicSlicesParameter.Value.Value;
                  double height = forecastModel.GetHeight() / forecastModel.GeographicSlicesParameter.Value.Value;

                  double regionSize = width * height;

                  var sliceCoordsUpperLeft = RegionHelper.GetCoordinates(upperLeft.Item1, upperLeft.Item2, forecastModel.GeographicSlicesParameter.Value.Value,
                   forecastModel.GetWidth(), forecastModel.GetHeight(), false);
                  var sliceCoordsLowerRight = RegionHelper.GetCoordinates(lowerRight.Item1, lowerRight.Item2, forecastModel.GeographicSlicesParameter.Value.Value,
                   forecastModel.GetWidth(), forecastModel.GetHeight(), false);

                  double sliceWidth = width;
                  double sliceHeight = height;

                  if (x == upperLeft.Item1) {
                    sliceWidth = width - ((posX - neighborhoodSize / 2) - sliceCoordsUpperLeft.Item1);
                  } else if (x == lowerRight.Item1) {
                    sliceWidth = (posX + neighborhoodSize / 2) - sliceCoordsLowerRight.Item1;
                  }

                  if (y == upperLeft.Item2) {
                    sliceHeight = height - ((posY - neighborhoodSize / 2) - sliceCoordsUpperLeft.Item2);
                  } else if (y == lowerRight.Item2) {
                    sliceHeight = (posY + neighborhoodSize / 2) - sliceCoordsLowerRight.Item2;
                  }

                  double sliceSize = sliceWidth * sliceHeight;

                  overlap = sliceSize / regionSize;
                }

                totalIntensity += intensity / overlap;
              }
            }
          }
        }

        //normalize acc. to time overlap
        double timeOverlap = 1.0;

        double timeSliceSize = forecastModel.GetHorizon() / forecastModel.TimeSlicesParameter.Value.Value; 
        if (t == m1) {
          double m1Time = RegionHelper.GetTime(m1 + 1, forecastModel.TimeSlicesParameter.Value.Value, forecastModel.GetHorizon(), false);
          double m1Overlap = m1Time - time;
          timeOverlap = m1Overlap / timeSliceSize;
        } else if (t == m2) {
          double m2Time = RegionHelper.GetTime(m2 - 1, forecastModel.TimeSlicesParameter.Value.Value, forecastModel.GetHorizon(), false);
          double m2Overlap = (time + waitingTime) - m2Time;
          timeOverlap = m2Overlap / timeSliceSize;
        }       

        probability *= Math.Exp(-totalIntensity * timeOverlap);
      }

      return 1.0 - probability;
    }

    //current wait information
    private IVRPEncoding currentSolution;

    private struct WaitInformation {
      public double PositionX { get; set; }
      public double PositionY { get; set; }
      public double StartTime { get; set; }
      public double EndTime { get; set; }
    }

    private Dictionary<Tour, List<WaitInformation>> waitInformation;

    private int GetWaitingVehicles(double time, double posX, double posY, HistoricalDataForecastModel forecastModel) {
      int waiting = 0;
      double neighborhoodSize = NeighborhoodSizeParameter.Value.Value * forecastModel.GetWidth();
    
      foreach (var tour in waitInformation.Keys) {
        int i = 0;
        while (i < waitInformation[tour].Count && waitInformation[tour][i].StartTime <= time) {
          if (waitInformation[tour][i].StartTime <= time && waitInformation[tour][i].EndTime > time) {
            if (Math.Abs(waitInformation[tour][i].PositionX - posX) <= neighborhoodSize / 2.0 &&
                Math.Abs(waitInformation[tour][i].PositionY - posY) <= neighborhoodSize / 2.0) {
              waiting++;
              break;
            }
          }

          i++;
        }
      }

      return waiting;
    }
  
    public override double GetWaitingTime(double time, DynPDPProblemInstance instance, IVRPEncoding solution, Tour tour, int stop) {
      if (solution != currentSolution) {
        currentSolution = solution;
        waitInformation = new Dictionary<Tour, List<WaitInformation>>();
      }

      var forecastModel = instance.Simulation.ForecastModelParameter.Value as HistoricalDataForecastModel;
      var scenario = instance.Scenario;
      double avgDist = GetAverageDistance(forecastModel.HistoricalDataParameter.Value, scenario.DistanceMeasure);

      int curCustomer = 0;
      if (stop > 0)
        curCustomer = tour.Stops[stop - 1];
      int nextCustomer = tour.Stops[stop];
      double curDist = instance.GetDistance(curCustomer, nextCustomer, solution);

      double minDist = MinDistanceParameter.Value.Value;  
      if (curDist <= avgDist * minDist) {
        return 0;
      } else {
        double maxWaitTime = MaxWaitTimeParameter.Value.Value * forecastModel.GetHorizon();
        double waitingTime = 
          Math.Min(maxWaitTime,
            WaitFirstWaitingStrategy.GetSlack(time, instance, solution, tour, stop));

        if (waitingTime > 0) {
          double curPosX, curPosY;
          if (curCustomer != 0) {
            curPosX = instance.GetCoordinates(curCustomer)[0];
            curPosY = instance.GetCoordinates(curCustomer)[1];
          } else {
            int depot = instance.GetDepot(nextCustomer, solution);
            curPosX = instance.DepotCoordinates[depot, 0];
            curPosY = instance.DepotCoordinates[depot, 1];
          }

          int maxWaitingVehicles = (int)Math.Round(MaxVehiclesParameter.Value.Value * instance.Vehicles.Value);
          int waitingVehicles = GetWaitingVehicles(time + instance.Simulation.SimulationTime, curPosX, curPosY, forecastModel);

          if (waitingVehicles <= maxWaitingVehicles) {
            double probability = GetProbability(time + instance.Simulation.SimulationTime, waitingTime, curPosX, curPosY, forecastModel);
            double threshold = ThresholdParameter.Value.Value;

            if (probability > threshold) {
              if (!waitInformation.ContainsKey(tour))
                waitInformation[tour] = new List<WaitInformation>();

              waitInformation[tour].Add(new WaitInformation() {
                PositionX = curPosX, PositionY = curPosY,
                StartTime = time + instance.Simulation.SimulationTime,
                EndTime = time + waitingTime + instance.Simulation.SimulationTime
              });

              return waitingTime;
            } else
              return 0;
          } else
            return 0;
        } else
          return 0;
      }
    }
  }
}
